@@ -10,9 +10,10 @@ import re
 
 from debian_linux import config
 from debian_linux.debian import PackageDescription, PackageRelation, \
-    PackageRelationEntry, PackageRelationGroup, VersionLinux
+    PackageRelationEntry, PackageRelationGroup, VersionLinux, \
+    restriction_requires_profile
 from debian_linux.gencontrol import Gencontrol as Base, merge_packages, \
-    iter_featuresets
+    iter_featuresets, iter_flavours, add_package_build_restriction
 from debian_linux.utils import Templates, read_control
 
 locale.setlocale(locale.LC_CTYPE, "C.UTF-8")
@@ -49,16 +50,22 @@ class Gencontrol(Base):
         },
         'packages': {
             'docs': config.SchemaItemBoolean(),
-            'headers-all': config.SchemaItemBoolean(),
             'installer': config.SchemaItemBoolean(),
             'libc-dev': config.SchemaItemBoolean(),
+            'meta': config.SchemaItemBoolean(),
             'tools-unversioned': config.SchemaItemBoolean(),
             'tools-versioned': config.SchemaItemBoolean(),
             'source': config.SchemaItemBoolean(),
         }
     }
 
-    def __init__(self, config_dirs=["debian/config"],
+    env_flags = [
+        ('DEBIAN_KERNEL_DISABLE_DEBUG', 'disable_debug', 'debug infos'),
+        ('DEBIAN_KERNEL_DISABLE_INSTALLER', 'disable_installer', 'installer modules'),
+        ('DEBIAN_KERNEL_DISABLE_SIGNED', 'disable_signed', 'signed code'),
+    ]
+
+    def __init__(self, config_dirs=["debian/config", "debian/config.local"],
                  template_dirs=["debian/templates"]):
         super(Gencontrol, self).__init__(
             config.ConfigCoreHierarchy(self.config_schema, config_dirs),
@@ -67,15 +74,29 @@ class Gencontrol(Base):
         self.process_changelog()
         self.config_dirs = config_dirs
 
+        for env, attr, desc in self.env_flags:
+            setattr(self, attr, False)
+            if os.getenv(env):
+                if self.changelog[0].distribution == 'UNRELEASED':
+                    import warnings
+                    warnings.warn(f'Disable {desc} on request ({env} set)')
+                    setattr(self, attr, True)
+                else:
+                    raise RuntimeError(
+                        f'Unable to disable {desc} in release build ({env} set)')
+
+        # In case installer packages are disabled through
+        # $DEBIAN_KERNEL_DISABLE_INSTALLER, update the config (that
+        # gets written to config.defines.dump) so that they will
+        # also be disabled in gencontrol_signed.py.
+        if self.disable_installer:
+            self.config.setdefault(('packages',), {})['installer'] = False
+
     def _setup_makeflags(self, names, makeflags, data):
         for src, dst, optional in names:
             if src in data or not optional:
                 makeflags[dst] = data[src]
 
-    def _substitute_file(self, template, vars, target, append=False):
-        with open(target, 'a' if append else 'w') as f:
-            f.write(self.substitute(self.templates[template], vars))
-
     def do_main_setup(self, vars, makeflags, extra):
         super(Gencontrol, self).do_main_setup(vars, makeflags, extra)
         makeflags.update({
@@ -84,25 +105,18 @@ class Gencontrol(Base):
             'ABINAME': self.abiname_version + self.abiname_part,
             'SOURCEVERSION': self.version.complete,
         })
-        makeflags['SOURCE_BASENAME'] = self.vars['source_basename']
+        makeflags['SOURCE_BASENAME'] = vars['source_basename']
+        makeflags['SOURCE_SUFFIX'] = vars['source_suffix']
 
         # Prepare to generate debian/tests/control
         self.tests_control = self.process_packages(
             self.templates['tests-control.main'], vars)
         self.tests_control_image = None
+        self.tests_control_headers = None
 
         self.installer_packages = {}
 
-        if os.getenv('DEBIAN_KERNEL_DISABLE_INSTALLER'):
-            if self.changelog[0].distribution == 'UNRELEASED':
-                import warnings
-                warnings.warn('Disable installer modules on request '
-                              '(DEBIAN_KERNEL_DISABLE_INSTALLER set)')
-            else:
-                raise RuntimeError(
-                    'Unable to disable installer modules in release build '
-                    '(DEBIAN_KERNEL_DISABLE_INSTALLER set)')
-        elif self.config.merge('packages').get('installer', True):
+        if not self.disable_installer and self.config.merge('packages').get('installer', True):
             # Add udebs using kernel-wedge
             kw_env = os.environ.copy()
             kw_env['KW_DEFCONFIG_DIR'] = 'debian/installer'
@@ -132,8 +146,11 @@ class Gencontrol(Base):
             # configuration errors before building linux-signed.
             build_signed = {}
             for arch in arches:
-                build_signed[arch] = self.config.merge('build', arch) \
-                                                .get('signed-code', False)
+                if not self.disable_signed:
+                    build_signed[arch] = self.config.merge('build', arch) \
+                                                    .get('signed-code', False)
+                else:
+                    build_signed[arch] = False
 
             for package in udeb_packages:
                 # kernel-wedge currently chokes on Build-Profiles so add it now
@@ -145,9 +162,10 @@ class Gencontrol(Base):
                     # convince debhelper and kernel-wedge to go
                     # part way to building them.
                     package['Build-Profiles'] = (
-                        '<pkg.linux.udeb-unsigned-test-build>')
+                        '<pkg.linux.udeb-unsigned-test-build !noudeb>')
                 else:
-                    package['Build-Profiles'] = '<!stage1 !pkg.linux.nokernel>'
+                    package['Build-Profiles'] = (
+                        '<!noudeb !stage1 !pkg.linux.nokernel>')
 
                 for arch in package['Architecture']:
                     self.installer_packages.setdefault(arch, []) \
@@ -170,23 +188,44 @@ class Gencontrol(Base):
 
     def do_main_packages(self, packages, vars, makeflags, extra):
         packages.extend(self.process_packages(
-            self.templates["control.main"], self.vars))
+            self.templates["control.main"], vars))
+
+        # Only build the metapackages if their names won't exactly match
+        # the packages they depend on
+        do_meta = self.config.merge('packages').get('meta', True) \
+            and vars['source_suffix'] != '-' + vars['version']
+
         if self.config.merge('packages').get('docs', True):
             packages.extend(self.process_packages(
-                self.templates["control.docs"], self.vars))
+                self.templates["control.docs"], vars))
+            if do_meta:
+                packages.extend(self.process_packages(
+                    self.templates["control.docs.meta"], vars))
+                self.substitute_debhelper_config(
+                    'docs.meta', vars,
+                    '%(source_basename)s-doc%(source_suffix)s' % vars)
         if self.config.merge('packages').get('tools-unversioned', True):
             packages.extend(self.process_packages(
-                self.templates["control.tools-unversioned"], self.vars))
+                self.templates["control.tools-unversioned"], vars))
         if self.config.merge('packages').get('tools-versioned', True):
             packages.extend(self.process_packages(
-                self.templates["control.tools-versioned"], self.vars))
+                self.templates["control.tools-versioned"], vars))
+            self.substitute_debhelper_config('perf', vars,
+                                              'linux-perf-%(version)s' % vars)
+            if do_meta:
+                packages.extend(self.process_packages(
+                    self.templates["control.tools-versioned.meta"], vars))
+                self.substitute_debhelper_config('perf.meta', vars,
+                                                  'linux-perf')
         if self.config.merge('packages').get('source', True):
             packages.extend(self.process_packages(
-                self.templates["control.sourcebin"], self.vars))
-
-        self._substitute_file('perf.lintian-overrides', self.vars,
-                              'debian/linux-perf-%s.lintian-overrides' %
-                              self.vars['version'])
+                self.templates["control.sourcebin"], vars))
+            if do_meta:
+                packages.extend(self.process_packages(
+                    self.templates["control.sourcebin.meta"], vars))
+                self.substitute_debhelper_config(
+                    'sourcebin.meta', vars,
+                    '%(source_basename)s-source%(source_suffix)s' % vars)
 
     def do_indep_featureset_setup(self, vars, makeflags, featureset, extra):
         makeflags['LOCALVERSION'] = vars['localversion']
@@ -238,36 +277,15 @@ class Gencontrol(Base):
 
     def do_arch_packages(self, packages, makefile, arch, vars, makeflags,
                          extra):
-        if self.version.linux_modifier is None:
-            try:
-                abiname_part = '-%s' % self.config['abi', arch]['abiname']
-            except KeyError:
-                abiname_part = self.abiname_part
-            makeflags['ABINAME'] = vars['abiname'] = \
-                self.abiname_version + abiname_part
-
-        build_signed = self.config.merge('build', arch) \
-                                  .get('signed-code', False)
-
-        # Some userland architectures require kernels from another
-        # (Debian) architecture, e.g. x32/amd64.
-        # And some derivatives don't need the headers-all packages
-        # for other reasons.
-        if self.config['base', arch].get('featuresets') and \
-           self.config.merge('packages').get('headers-all', True):
-            headers_arch = self.templates["control.headers.arch"]
-            packages_headers_arch = self.process_packages(headers_arch, vars)
-            packages_headers_arch[-1]['Depends'].extend(PackageRelation())
-            extra['headers_arch_depends'] = (
-                packages_headers_arch[-1]['Depends'])
+        if not self.disable_signed:
+            build_signed = self.config.merge('build', arch) \
+                                      .get('signed-code', False)
         else:
-            packages_headers_arch = []
+            build_signed = False
 
         if self.config.merge('packages').get('libc-dev', True):
             libc_dev = self.templates["control.libc-dev"]
-            packages_headers_arch[0:0] = self.process_packages(libc_dev, {})
-
-        merge_packages(packages, packages_headers_arch, arch)
+            merge_packages(packages, self.process_packages(libc_dev, {}), arch)
 
         if self.config['base', arch].get('featuresets') and \
            self.config.merge('packages').get('source', True):
@@ -318,12 +336,28 @@ class Gencontrol(Base):
         vars['localversion_headers'] = vars['localversion']
         makeflags['LOCALVERSION_HEADERS'] = vars['localversion_headers']
 
+        self.default_flavour = self.config.merge('base', arch, featureset) \
+                                          .get('default-flavour')
+        if self.default_flavour is not None:
+            if featureset != 'none':
+                raise RuntimeError("default-flavour set for %s %s,"
+                                   " but must only be set for featureset none"
+                                   % (arch, featureset))
+            if self.default_flavour \
+               not in iter_flavours(self.config, arch, featureset):
+                raise RuntimeError("default-flavour %s for %s %s does not exist"
+                                   % (self.default_flavour, arch, featureset))
+
+        self.quick_flavour = self.config.merge('base', arch, featureset) \
+                                        .get('quick-flavour')
+
     flavour_makeflags_base = (
         ('compiler', 'COMPILER', False),
         ('compiler-filename', 'COMPILER', True),
         ('kernel-arch', 'KERNEL_ARCH', False),
-        ('cflags', 'CFLAGS_KERNEL', True),
+        ('cflags', 'KCFLAGS', True),
         ('override-host-type', 'OVERRIDE_HOST_TYPE', True),
+        ('cross-compile-compat', 'CROSS_COMPILE_COMPAT', True),
     )
 
     flavour_makeflags_build = (
@@ -347,6 +381,7 @@ class Gencontrol(Base):
                                                flavour)
         config_image = self.config.merge('image', arch, featureset, flavour)
 
+        vars['flavour'] = vars['localversion'][1:]
         vars['class'] = config_description['hardware']
         vars['longclass'] = (config_description.get('hardware-long')
                              or vars['class'])
@@ -371,6 +406,7 @@ class Gencontrol(Base):
         headers = self.templates["control.headers"]
         assert len(headers) == 1
 
+        do_meta = self.config.merge('packages').get('meta', True)
         config_entry_base = self.config.merge('base', arch, featureset,
                                               flavour)
         config_entry_build = self.config.merge('build', arch, featureset,
@@ -393,8 +429,10 @@ class Gencontrol(Base):
             self.substitute(config_entry_relations.get('headers%' + compiler)
                             or config_entry_relations.get(compiler), vars))
         relations_compiler_headers = PackageRelation(
-            PackageRelationGroup(entry for entry in group
-                                 if 'cross' not in entry.restrictions)
+            PackageRelationGroup(
+                entry for entry in group
+                if not restriction_requires_profile(entry.restrictions,
+                                                    'cross'))
             for group in relations_compiler_headers)
         for group in relations_compiler_headers:
             for entry in group:
@@ -457,7 +495,10 @@ class Gencontrol(Base):
 
         packages_own = []
 
-        build_signed = config_entry_build.get('signed-code')
+        if not self.disable_signed:
+            build_signed = config_entry_build.get('signed-code')
+        else:
+            build_signed = False
 
         image = self.templates[build_signed and "control.image-unsigned"
                                or "control.image"]
@@ -476,26 +517,58 @@ class Gencontrol(Base):
             extra['headers_arch_depends'].append('%s (= ${binary:Version})' %
                                                  packages_own[-1]['Package'])
 
+        # The image meta-packages will depend on signed linux-image
+        # packages where applicable, so should be built from the
+        # signed source packages The header meta-packages will also be
+        # built along with the signed packages, to create a dependency
+        # relationship that ensures src:linux and src:linux-signed-*
+        # transition to testing together.
+        if do_meta and not build_signed:
+            packages_meta = self.process_packages(
+                self.templates['control.image.meta'], vars)
+            assert len(packages_meta) == 1
+            packages_meta += self.process_packages(
+                self.templates['control.headers.meta'], vars)
+            assert len(packages_meta) == 2
+
+            if flavour == self.default_flavour \
+               and not self.vars['source_suffix']:
+                packages_meta[0].setdefault('Provides', PackageRelation()) \
+                                .append('linux-image-generic')
+                packages_meta[1].setdefault('Provides', PackageRelation()) \
+                                .append('linux-headers-generic')
+
+            packages_own.extend(packages_meta)
+
+            self.substitute_debhelper_config(
+                "image.meta", vars,
+                "linux-image%(localversion)s" % vars)
+            self.substitute_debhelper_config(
+                'headers.meta', vars,
+                'linux-headers%(localversion)s' % vars)
+
         if config_entry_build.get('vdso', False):
             makeflags['VDSO'] = True
 
-        build_debug = config_entry_build.get('debug-info')
-
-        if os.getenv('DEBIAN_KERNEL_DISABLE_DEBUG'):
-            if self.changelog[0].distribution == 'UNRELEASED':
-                import warnings
-                warnings.warn('Disable debug infos on request '
-                              '(DEBIAN_KERNEL_DISABLE_DEBUG set)')
-                build_debug = False
-            else:
-                raise RuntimeError(
-                    'Unable to disable debug infos in release build '
-                    '(DEBIAN_KERNEL_DISABLE_DEBUG set)')
+        if not self.disable_debug:
+            build_debug = config_entry_build.get('debug-info')
+        else:
+            build_debug = False
 
         if build_debug:
-            makeflags['DEBUG'] = True
             packages_own.extend(self.process_packages(
                 self.templates['control.image-dbg'], vars))
+            if do_meta:
+                packages_own.extend(self.process_packages(
+                    self.templates["control.image-dbg.meta"], vars))
+                self.substitute_debhelper_config(
+                    'image-dbg.meta', vars,
+                    'linux-image%(localversion)s-dbg' % vars)
+
+        # In a quick build, only build the quick flavour (if any).
+        if flavour != self.quick_flavour:
+            for package in packages_own:
+                add_package_build_restriction(package, '!pkg.linux.quick')
 
         merge_packages(packages, packages_own, arch)
 
@@ -511,6 +584,16 @@ class Gencontrol(Base):
             self.tests_control_image = tests_control
             self.tests_control.append(tests_control)
 
+        if flavour == (self.quick_flavour or self.default_flavour):
+            if not self.tests_control_headers:
+                self.tests_control_headers = self.process_package(
+                    self.templates['tests-control.headers'][0], vars)
+                self.tests_control.append(self.tests_control_headers)
+            self.tests_control_headers['Architecture'].add(arch)
+            self.tests_control_headers['Depends'].append(
+                PackageRelationGroup(package_headers['Package'],
+                                     override_arches=(arch,)))
+
         def get_config(*entry_name):
             entry_real = ('image',) + entry_name
             entry = self.config.get(entry_real, None)
@@ -561,14 +644,9 @@ class Gencontrol(Base):
                                     arch, featureset, flavour))
         makeflags['KCONFIG'] = ' '.join(kconfig)
         makeflags['KCONFIG_OPTIONS'] = ''
-        if build_debug:
-            makeflags['KCONFIG_OPTIONS'] += ' -o DEBUG_INFO=y'
-        if build_signed:
-            makeflags['KCONFIG_OPTIONS'] += ' -o MODULE_SIG=y'
         # Add "salt" to fix #872263
-        makeflags['KCONFIG_OPTIONS'] += (' -o "BUILD_SALT=\\"%s%s\\""' %
-                                         (vars['abiname'],
-                                          vars['localversion']))
+        makeflags['KCONFIG_OPTIONS'] += \
+            ' -o "BUILD_SALT=\\"%(abiname)s%(localversion)s\\""' % vars
 
         cmds_binary_arch = ["$(MAKE) -f debian/rules.real binary-arch-flavour "
                             "%s" %
@@ -590,33 +668,67 @@ class Gencontrol(Base):
                      cmds=["$(MAKE) -f debian/rules.real %s %s" %
                            (merged_config, makeflags)])
 
-        # Substitute kernel version etc. into maintainer scripts,
-        # translations and lintian overrides
-        self._substitute_file('headers.postinst', vars,
-                              'debian/linux-headers-%s%s.postinst' %
-                              (vars['abiname'], vars['localversion']))
-        for name in ['postinst', 'postrm', 'preinst', 'prerm']:
-            self._substitute_file('image.%s' % name, vars,
-                                  'debian/%s.%s' %
-                                  (image_main['Package'], name))
+        self.substitute_debhelper_config(
+            'headers', vars,
+            'linux-headers-%(abiname)s%(localversion)s' % vars)
+        self.substitute_debhelper_config('image', vars, image_main['Package'])
         if build_debug:
-            debug_lintian_over = (
-                'debian/linux-image-%s%s-dbg.lintian-overrides' %
-                (vars['abiname'], vars['localversion']))
-            self._substitute_file('image-dbg.lintian-overrides', vars,
-                                  debug_lintian_over)
-            os.chmod(debug_lintian_over, 0o755)
+            self.substitute_debhelper_config(
+                'image-dbg', vars,
+                'linux-image-%(abiname)s%(localversion)s-dbg' % vars)
+
+    # Generate a unique kernel ABI name suffix following the old
+    # (pre-trixie) format
+    def _generate_abiname_part(self):
+        # Start with 35 for 5.10.237-1 (the last manually set ABI
+        # version) and add 1 for every non-backport changelog entry
+        # after that.  Also count backport entries since the last
+        # non-backport, in case we need further disambiguation.
+        abiname_major = 0
+        backport_count = 0
+        target_release = None
+        for entry in self.changelog:
+            if '~' in entry.version.revision:
+                if target_release is None:
+                    target_release = entry.version.revision.split('u')[0]
+                if abiname_major == 0:
+                    # Only increase for backport entries for the target release
+                    if target_release in entry.version.revision:
+                        backport_count += 1
+            elif str(entry.version) == '5.10.237-1':
+                abiname_major += 35
+                break
+            else:
+                abiname_major += 1
+
+        # Select an appropriate prefix for the target release
+        for release_name, release_prefix in [
+                ('UNRELEASED', ''),
+                ('bullseye',   ''),
+                ('buster',     '0.deb10.'),
+                ('stretch',    '0.deb9.'),
+        ]:
+            if self.changelog[0].distribution.startswith(release_name):
+                break
+        else:
+            assert False, f'cannot generate ABI name for {self.changelog[0].distribution}'
+
+        # Paste that all together
+        result = f'-{release_prefix}{abiname_major}'
+        if backport_count > 1:
+            result += f'.{backport_count - 1}'
+        return result
 
     def process_changelog(self):
         version = self.version = self.changelog[0].version
-        if self.version.linux_modifier is not None:
-            self.abiname_part = ''
-        else:
+        try:
             self.abiname_part = '-%s' % self.config['abi', ]['abiname']
+        except KeyError:
+            self.abiname_part = self._generate_abiname_part()
         # We need to keep at least three version components to avoid
         # userland breakage (e.g. #742226, #745984).
         self.abiname_version = re.sub(r'^(\d+\.\d+)(?=-|$)', r'\1.0',
-                                      self.version.linux_upstream)
+                                      self.version.linux_version)
         self.vars = {
             'upstreamversion': self.version.linux_upstream,
             'version': self.version.linux_version,
@@ -626,6 +738,8 @@ class Gencontrol(Base):
             'source_package': self.changelog[0].source,
             'abiname': self.abiname_version + self.abiname_part,
         }
+        self.vars['source_suffix'] = \
+            self.changelog[0].source[len(self.vars['source_basename']):]
         self.config['version', ] = {'source': self.version.complete,
                                     'upstream': self.version.linux_upstream,
                                     'abiname_base': self.abiname_version,
